{**********************************************************************
 Package pl_FBDataSet
 this unit is part of CodeTyphon Studio (http://www.pilotlogic.com/)
***********************************************************************}

unit allFBDataSetreg;

  {$H+}
  {$I fb_define.inc}
interface

uses
  Classes, SysUtils, typinfo, Forms, LCLType, Controls,
  LazarusPackageIntf,PropEdits,componenteditors,
  DB,DBPropEdits,
  uib,  uiblib,
  jvuibdatabaseedit, jvuibtransactionedit,
  UIBStoredProc, FBCustomDataSet,
  fbcustomdatasetpropeditorl,
  metadatasqlgenerator;

type

  { TUIBDatabaseEditor }

  TUIBDatabaseEditor = class(TComponentEditor)
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  { TUIBTransactionEditor }

  TUIBTransactionEditor = class(TComponentEditor)
  public
    procedure ExecuteVerb(Index: Integer); override;
    function GetVerb(Index: Integer): string; override;
    function GetVerbCount: Integer; override;
  end;

  { TUIBDatabaseDBNameProperty }

  TUIBDatabaseDBNameProperty = class(TStringPropertyEditor)
  public
    function GetAttributes: TPropertyAttributes; override;
    procedure Edit; override;
  end;

  { TUIBDatabaseRoleNameProperty }

  TUIBDatabaseRoleNameProperty = class(TFieldProperty)
  public
    procedure FillValues(const Values: TStringList); override;
    procedure SetValue(const NewValue: ansistring); override;
  end;

procedure Register;

implementation

{ TUIBDatabaseEditor }

procedure TUIBDatabaseEditor.ExecuteVerb(Index: Integer);
begin
  with TUIBDatabaseEditForm2.Create(Application) do
  try
    Database := TUIBDataBase(Component);
    if ShowModal = mrOk then
      inherited Designer.Modified;
  finally
    Free;
  end;
end;

function TUIBDatabaseEditor.GetVerb(Index: Integer): string;
begin
  Result := 'Database Editor ...';
end;

function TUIBDatabaseEditor.GetVerbCount: Integer;
begin
  Result := 1;
end;

{ TUIBTransactionEditor }

procedure TUIBTransactionEditor.ExecuteVerb(Index: Integer);
begin
  with TUIBTransactionEditForm2.Create(Application) do
  try
    Transaction := TUIBTransaction(Component);
    if ShowModal = mrOk then
      inherited Designer.Modified;
  finally
    Free;
  end;
end;

function TUIBTransactionEditor.GetVerb(Index: Integer): string;
begin
  Result := 'Transaction Editor ...';
end;

function TUIBTransactionEditor.GetVerbCount: Integer;
begin
  Result:=1;
end;

{ TUIBDatabaseDBNameProperty }

function TUIBDatabaseDBNameProperty.GetAttributes: TPropertyAttributes;
begin
  Result:=inherited GetAttributes + [paDialog];
end;

procedure TUIBDatabaseDBNameProperty.Edit;
var
  UIBDataBase: TUIBDataBase;
begin
  UIBDataBase := GetComponent(0) as TUIBDataBase;
  with TUIBDatabaseEditForm2.Create(Application) do
  try
    Database := TUIBDataBase(UIBDataBase);
    if ShowModal = mrOk then
      Modified;
  finally
    Free;
  end;
end;

{ TUIBDatabaseRoleNameProperty }
const
  sRoleList =
    'select rdb$roles.rdb$role_name from rdb$roles order by rdb$roles.rdb$role_name';

procedure TUIBDatabaseRoleNameProperty.FillValues(const Values: TStringList);
var
  D:TUIBDataBase;
  Q:TUIBQuery;
  Tr:TUIBTransaction;
begin
  D:=TUIBDataBase(GetComponent(0));
  if not (D is TUIBDataBase) then exit;
  if not D.Connected then exit;

  Tr:=TUIBTransaction.Create(nil);
  Tr.DataBase:=D;
  Tr.Options:=[tpRead, tpReadCommitted, tpRecVersion];
  Tr.StartTransaction;
  Q:=TUIBQuery.Create(nil);
  try
    Q.DataBase:=D;
    Q.Transaction:=Tr;
    Q.SQL.Text:=sRoleList;
    Q.Open;
    Values.Clear;
    while not Q.Eof do
    begin
      Values.Add(Trim(Q.Fields.AsString[0]));
      Q.Next;
    end;
    Q.Close;
  finally
    Q.Free;
    Tr.Free;
  end;
end;

procedure TUIBDatabaseRoleNameProperty.SetValue(const NewValue: ansistring);
begin
  try
    inherited SetValue(NewValue);
  finally
  end;
end;

//=======================================================

procedure Register;
begin

  RegisterComponents('UIB', [TUIBStoredProc,
                             TFBDataSet,
                             TMetaDataSQLGenerator]);

  RegisterClass(TFBBlobField);
  RegisterClass(TFBAnsiMemoField);
  RegisterClass(TFBAnsiField);
  RegisterNoIcon([TFBAnsiMemoField, TFBAnsiField, TFBBlobField, TFBLargeintField]);

  RegisterComponentEditor(TUIBDataBase, TUIBDatabaseEditor);
  RegisterComponentEditor(TUIBTransaction, TUIBTransactionEditor);

  RegisterPropertyEditor(TypeInfo(string), TUIBDataBase, 'DatabaseName', TUIBDatabaseDBNameProperty);
  RegisterPropertyEditor(TypeInfo(string), TUIBDataBase, 'Role', TUIBDatabaseRoleNameProperty);

  {$IFDEF FBSQLEditor}
  RegisterPropertyEditor(TypeInfo(TStrings), TFBDataSet, 'SQLSelect', TFBDataSetSQLProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TFBDataSet, 'SQLRefresh', TFBDataSetSQLProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TFBDataSet, 'SQLEdit', TFBDataSetSQLProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TFBDataSet, 'SQLDelete', TFBDataSetSQLProperty);
  RegisterPropertyEditor(TypeInfo(TStrings), TFBDataSet, 'SQLInsert', TFBDataSetSQLProperty); {do not localize}
  {$ENDIF}
  RegisterPropertyEditor(TypeInfo(TAutoUpdateOptions), TFBDataSet, 'AutoUpdateOptions', TAutoUpdateOptionsProperty); {do not localize}

  RegisterComponentEditor(TFBDataSet, TFBDataSetEditor);

  RegisterPropertyEditor(TypeInfo(string), TUIBStoredProc, 'StoredProcName', TUIBStoredProcNameProperty);

end;

end.

